#pragma once

/* Genode includes */
#include <util/reconstructible.h>

/* local includes */
#include "context.h"
#include "types.h"


namespace State_machine
{
	//using namespace Genode;
	using Genode::Xml_node;
	using Genode::size_t;

	class Provider;
	class Standard_provider;
	class Buffered_provider;
	class Provider_factory;
}


class State_machine::Provider
{
	private:

		Context                         &_context;
		Xml_node                  const  _config;
		Signal_context_capability        _sigh { };

	protected:

		Context                   &context()       { return _context; }
		Allocator                 &alloc()         { return _context.alloc(); };
		Xml_node           const  &config() const  { return _config; }

		void notify_sigh()
		{
			if (_sigh.valid()) {
				Signal_transmitter(_sigh).submit();
			}
		}

	public:

		virtual char const *data()    const = 0;
		virtual size_t      length()  const = 0;
		virtual void        compute()       = 0;

		virtual void sigh(Signal_context_capability sigh)
		{
			_sigh = sigh;
		}

		Provider(Context &context, Xml_node const &config) :
			_context { context },
			_config  { config }
		{ }

		virtual ~Provider() = default;
};


class State_machine::Standard_provider : public State_machine::Provider
{
	private:

		static constexpr char const *_dummy { "\0" };

		struct Data
		{
			char   const  *data;
			size_t         length;
		};

		Data  _data  { .data = _dummy,
		               .length = strlen(_dummy) };

	protected:

		void update_data(char const *data, size_t length)
		{
			_data.data   = data;
			_data.length = length;
		}

	public:

		char const *data() const override
		{
			return _data.data;
		}

		size_t length() const override
		{
			return _data.length;
		}

		Standard_provider(Context &env, Xml_node const &config) :
			Provider  { env, config }
		{ }
};


class State_machine::Buffered_provider : public State_machine::Provider
{
	private:

		class Buffer
		{
			private:
				Allocator  &_alloc;
				size_t      _capacity;
				size_t      _length;
				char       *_data;

				size_t _compute_capacity(size_t length)
				{
					constexpr size_t const max_length { static_cast<size_t>(~0ull / 2ull) };

					if (length > max_length)
						throw Length_too_large { };

					size_t capacity = 8;

					while (capacity < length) {
						capacity *= 2;
					}

					return capacity;
				}

			public:
				size_t       capacity()       { return _capacity; }
				size_t       length()   const { return _length; }
				char const  *data()     const { return _data; }

				void append(char const *data, size_t length)
				{
					if (length + _length > _capacity) {
						_capacity = _compute_capacity(length + _length);
						char *new_data { static_cast<char*>(_alloc.alloc(_capacity)) };
						memcpy(new_data, _data, _length);
						destroy(&_alloc, _data);
						_data = new_data;
					}

					memcpy(_data + _length, data, length);
					_length += length;
				}

				void replace(char const *data, size_t length)
				{
					if (length > _capacity) {
						destroy(&_alloc, _data);
						_capacity = _compute_capacity(length);
						_data = static_cast<char*>(_alloc.alloc(_capacity));
					}

					memcpy(_data, data, length);
					_length = length;
				}

				Buffer(Allocator &alloc, size_t capacity) :
					_alloc { alloc },
					_capacity { capacity },
					_length { 0 },
					_data { static_cast<char*>(_alloc.alloc(capacity)) }
				{ }

				~Buffer()
				{
					destroy(&_alloc, _data);
				}

				Buffer(const Buffer&)            = delete;
				Buffer &operator=(const Buffer&) = delete;
		};

		Buffer _buffer { alloc(), 8 };

	protected:

		void set(char const *data, size_t length)
		{
			_buffer.replace(data, length);
		}

		void append(char const *data, size_t length)
		{
			_buffer.append(data, length);
		}

	public:
		char const *data()   const override { return _buffer.data(); }
		size_t      length() const override { return _buffer.length(); }

		Buffered_provider(Context &env, Xml_node const &config) :
			Provider { env, config }
		{ }

		Buffered_provider(const Buffered_provider&)            = delete;
		Buffered_provider &operator=(const Buffered_provider&) = delete;
};


class State_machine::Provider_factory
{
	public:
		virtual State_machine::Provider &create(Context &env, Xml_node const &config) = 0;

		virtual ~Provider_factory() = default;
};

